import numpy as np
from scipy.integrate import cumtrapz

import mfNRcatpy.base as bs

def EOS_masses():
    D = {'BHBlp':{'maxm_TOV':2.100904, 'maxm_rigrot':2.549901, 'maxm_bar_TOV':2.462751,'maxm_bar_rigrot':2.961749,'ref':'Banik et al., Astrophys.J.Suppl. 214 (2014) no.2, 22'},
         'DD2':{'maxm_TOV':2.422658, 'maxm_rigrot':2.955318, 'maxm_bar_TOV':2.926458,'maxm_bar_rigrot':3.521266,'ref':'Typel et al., Phys.Rev. C81 (2010) 015803'},
         'LS220':{'maxm_TOV':2.043796, 'maxm_rigrot':2.451828, 'maxm_bar_TOV':2.429123,'maxm_bar_rigrot':2.881334,'ref':'Lattimer and Swesty, Nucl.Phys. A535 (1991) 331-376'},
         'SFHo':{'maxm_TOV':2.058882, 'maxm_rigrot':2.473178, 'maxm_bar_TOV':2.457619,'maxm_bar_rigrot':2.912048,'ref':'Steiner and Hempel and Fischer, Astrophys.J. 774 (2013) 17'},
         '2B':{'maxm_TOV':1.783269, 'maxm_rigrot':2.162003, 'maxm_bar_TOV':2.136538,'maxm_bar_rigrot':2.580824,'ref':'Read et al., Phys.Rev. D79 (2009) 124032'},
         '2H':{'maxm_TOV':2.834886, 'maxm_rigrot':3.461828, 'maxm_bar_TOV':3.409756,'maxm_bar_rigrot':4.145367,'ref':'Read et al., Phys.Rev. D79 (2009) 124032'},
         'ALF2':{'maxm_TOV':1.99092, 'maxm_rigrot':2.510254, 'maxm_bar_TOV':2.319515,'maxm_bar_rigrot':2.928461,'ref':'Read et al., Phys. Rev. D79 (2009) 124032'},
         'ENG':{'maxm_TOV':2.25052, 'maxm_rigrot':2.759074, 'maxm_bar_TOV':2.734237,'maxm_bar_rigrot':3.33214,'ref':'Read et al., Phys. Rev. D79 (2009) 124032'},
         'H4':{'maxm_TOV':2.028068, 'maxm_rigrot':2.476984, 'maxm_bar_TOV':2.332485,'maxm_bar_rigrot':2.855665,'ref':'Read et al., Phys. Rev. D79 (2009) 124032'},
         'MPA1':{'maxm_TOV':2.470327, 'maxm_rigrot':3.065678, 'maxm_bar_TOV':3.035605,'maxm_bar_rigrot':3.74128,'ref':'Read et al., Phys.Rev. D79 (2009) 124032'},
         'MS1':{'maxm_TOV':2.769299, 'maxm_rigrot':3.427183, 'maxm_bar_TOV':3.348443,'maxm_bar_rigrot':4.127858,'ref':'Read et al., Phys.Rev. D79 (2009) 124032'},
         'MS1b':{'maxm_TOV':2.76283, 'maxm_rigrot':3.439019, 'maxm_bar_TOV':3.353737,'maxm_bar_rigrot':4.155828,'ref':'Read et al., Phys.Rev. D79 (2009) 124032'},
         'SLy':{'maxm_TOV':2.060435, 'maxm_rigrot':2.507342, 'maxm_bar_TOV':2.46085,'maxm_bar_rigrot':2.985524,'ref':'Read et al., Phys.Rev. D79 (2009) 124032'}}
    return D

def chi_eff(m1,m2, x1, x2):
    '''
    equation 4 in core paper
    https://arxiv.org/abs/1806.01625
    
    m1 & m2 masses in solar masses
    x1 & x2 spin magnitudes
    
    The label 1 is reserved for the most massive object
    
    '''
    M = m1+m2
    chi = (m1*x1)/M + (m2*x2)/M - (38/113)*(m1*m2/(M**2))*(x1+x2)
    return chi

def lambda_tilde(m1, m2, l1, l2):
    '''
    eq 5 core paper
    https://arxiv.org/abs/1806.01625
    
    m1 & m2 masses in solar masses
    l1 & l2 spin magnitudes
    
    The label 1 is reserved for the most massive object
    '''
    L = (16/13) *( ((m1+(12*m2))*m1*l1)/((m1+m2)**5) + ((m2+(12*m1))*m2*l2)/((m2+m1)**5) )
    
    return L
def lambda_from_compactness(m, R):
    '''
    universal relation taken from 
    https://journals.aps.org/prd/pdf/10.1103/PhysRevD.102.043023
    '''
    
    C = m/R
    pass
    
    
    

def energylm(l,m , hlm, t, r):
    '''
    uses equation(10) from https://arxiv.org/abs/1311.4443
    
    input:    
        l,m: int,int. Indices from spin weighted spherical harmonics expansion
        hlm: complex array. Spherical harmonics terms coming from integrating twice the weyl scalar ones.
        t:  time array. Must be in seconds
        r:  float. Distance to the source, must be in meters
        
    output:
        e_MSUN: time domain object.
            e_MSUN.td: time array. given in seconds
            e_MSUN.data: data array. energy radiated given in solar masses
    
        luminosity: time domain object.
            luminosity.td: time array. given in seconds
            luminosity.data: data array. dimensionless
    
    '''
    MSUN_SEC = 4.92e-6
    meter_in_sec = (1/3e8)
    r=r*meter_in_sec
    
    integrand = np.abs(r*np.gradient(hlm,t,edge_order=2))**2 
    
    e_MSUN =  bs.time_domain(t,(1/(16*np.pi))*cumtrapz(integrand, t, initial=0)/MSUN_SEC)
    luminosity = bs.time_domain(t, integrand)
    
    return e_MSUN, luminosity

def angmom_Z_lm(l,m, hlm, t,r):
    '''
    uses equation(11) from https://arxiv.org/abs/1311.4443 with a minor correction
    i.e. here i take the imaginary part of the integrand as suggested by
    https://arxiv.org/abs/0707.4654 in equation (49)
    input:    
        l,m: int,int. Indices from spin weighted spherical harmonics expansion
        hlm: complex array. Spherical harmonics terms coming from integrating twice the weyl scalar ones.
        t:  time array. Must be in seconds
        r:  float. Distance to the source, must be in meters
        
    output:
        e_MSUN: time domain object.
            e_MSUN.td: time array. given in seconds
            e_MSUN.data: data array. energy radiated given in solar masses
    
        luminosity: time domain object.
            luminosity.td: time array. given in seconds
            luminosity.data: data array. dimensionless
    
    '''

    MSUN_SEC = 4.92e-6
    meter_in_sec = (1/3e8)
    r=r*meter_in_sec
    
    integrand = -(m*(r**2)* hlm * np.gradient(np.conj(hlm), t,edge_order=2)).imag
    
    Jz_MSUNsq = bs.time_domain(t, (1/(16*np.pi))*cumtrapz(integrand, t, initial=0)/MSUN_SEC**2)
    torquez_msun = bs.time_domain(t, integrand/MSUN_SEC)
    
    return Jz_MSUNsq, torquez_msun



import numpy as np
from scipy import interpolate
from scipy import signal
from numba import njit, prange

import matplotlib.pyplot as plt
import os

import mfNRcatpy.base as bs
import mfNRcatpy.utils as ut
#from mfNRcatpy.core_wf import *
#from mfNRcatpy.lal_wf import *




@njit(parallel=True, fastmath=True)
def integral(df, tf, new_inv_PSD, F, delta_f, tau):
    '''
    to use this you better install numba and SVML

    conda install numba
    conda install -c numba icc_rt

    '''

    templ_dot_templ = 4 * (np.sum(tf * tf.conjugate() * new_inv_PSD)) * delta_f 
    N = np.sqrt(templ_dot_templ)

    # shifting inner product
    L = []
    for i in prange(tau.size):
    #for t in tau:
        #S = 4*(df * tf.conjugate()) * new_inv_PSD * np.exp(1j*2*np.pi*F*t) * delta_f
        S = 4*(df * tf.conjugate()) * new_inv_PSD * np.exp(1j*2*np.pi*F*tau[i]) * delta_f

        rho = S/N
        L.append(np.sum(rho))

    L = np.array(L)
    return L

def extend_t(da,N, method=None):
    '''
    this function extends an array of floats
    symmetrically to left and right with N new 
    values based on the sampling of the original
    array da e.g.
    
    a = np.array([1,2,3,4,5])
    extend_t(a,4)
    '''
    delta = abs(da[1]-da[0])
    new_da = da
    if method=='loop':
        for i in range(1,N+1):
            # insert the left value
            left = (-delta) + new_da[0]
            new_da = np.flip(new_da)
            new_da = np.append(new_da, left)
            new_da = np.flip(new_da)
            # insert the right value
            right = (delta) + new_da[-1]
            new_da = np.append(new_da, right)
    elif method is None or method=='vector':
        arr = np.linspace(1,N, N)*delta
        left = da[0]-np.flip(arr)
        right = da[-1]+arr
        new_da = np.concatenate((left,da,right))
        
    return new_da


def tau_shift_gen(A, dt):
    '''
    REMARK: You might want to multiply the output
    by some dt, to scale correctly the time shift 
    domain.
    
    
    This function generates the timeshift 
    domain for convolutions or cross correlations
    
    Since this depends on if the array's lenght is
    odd or even, this uses the convention in which,
    
    if the input array has even lenght (it has 2 center values),
    then it picks the one which is on the right, in order 
    to get more negative values before the zero value
    
    Arrays with odd leghts do not represent 
    any confusion since they have just 1 center
    
    Input
    A : cross correlation or convolution array
    
    Output:
    B : time shift domain for a cross correlation
    
    example:
    
    a = np.arange(8)
    b = np.arange(9)

    print('Array with even lenght {}'.format(tau_shift_gen(a)))
    print('')
    print('Array with odd lenght {}'.format(tau_shift_gen(b)))
    
    '''
    N = A.size
    if N%2==0: #par
        B = np.arange(-(N//2),(N//2))
    else:      #impar
        B = np.arange(-(N//2),(N//2)+1)
    return B * dt



class matched_filtering(object):
    '''
    Here I'm supposed to have already decided 
    if i want the full strain or just hp, hc
    
    template  and waveform need to be converted
    to time domain objects beforehand
    
    '''
    
    def __init__(self, td_wf, td_templ, ASD):
        
        self.wf = td_wf
        self.templ = td_templ
        self.ASD = ASD
        
    def cross_correlation(self,max_templ_dur=None):        
        data_td, template_td = self.length_matching(max_templ_dur)
        key = 'same'
        dt = data_td.dt
        c =  signal.correlate(data_td.data, template_td.data, mode=key)*data_td.dt
        
        tau0 = np.arange(c.size)
        tau = tau_shift_gen(c, dt)
        
        return bs.time_domain(tau, c)
        
        
        
    
    def optimal_SNR(self, N=None, cut_sc=None):
        a = self.templ
        self.templ = self.wf
        s_opt = self.recovered_SNR3(N=N, cut_sc=cut_sc)
        self.templ = a
        return s_opt
    


    
    def recovered_SNR3(self, N=None, cut_sc=None):
        
        w = self.wf
        t = self.templ
        sc = self.ASD
        
        if N is None:
            N = 5

        if cut_sc is None:
            sc = sc.cut([50, 4500])
        else:
            sc = sc.cut(cut_sc)



        Mf = matched_filtering(w, t, sc)
        # extend the arrays and iterpolate to evaluate the same samples
        data, templ = Mf.length_matching2(N=N)
        delta_t = data.dt

        # to frequency domain using ffts    
        # but reorder them to get zero freq in the center
        # in this way is easier to debug
        F = np.fft.fftshift(np.fft.fftfreq(data.data.size, delta_t))
        delta_f = F[1]-F[0]

        df, tf = [np.fft.fftshift(np.fft.fft(data.data)*delta_t),
                  np.fft.fftshift(np.fft.fft(templ.data)*delta_t)]

        # define the inverse power spectrum of the detector    
        new_inv_PSD = sc.inv_psd_extension(F).PSD_flip_mirror(F).data    
        data_dot_templ = (df * tf.conjugate()) * new_inv_PSD    
        mask = F<0
        data_dot_templ[mask]=0

        # pull the zero frequency back to the first array position
        data_dot_templ = np.fft.ifftshift(data_dot_templ)
        # use ifft
        data_dot_templ = 4* (np.fft.ifft(data_dot_templ)*data_dot_templ.size) * delta_f
        S = data_dot_templ

        # 2. denominator --> Calculate the norm of the template

        templ_dot_templ = tf* tf.conjugate() * new_inv_PSD
        templ_dot_templ[mask] = 0
        templ_dot_templ =  4*(np.sum(templ_dot_templ) * delta_f)

        N = np.sqrt(templ_dot_templ)
        rho = np.fft.fftshift(S/N)
        tau = tau_shift_gen(data.data, delta_t)
        return bs.time_domain(tau, rho)
        

    def length_matching2(self, N=None):

        A = (self.wf).data
        da = (self.wf).td

        B = (self.templ).data
        db = (self.templ).td

        if N is None:
            N=50

        dur_a = da[-1]-da[0]
        dur_b = db[-1]-db[0]

        delta_t = da[1]-da[0]

        ti, tf =  np.take(np.sort([da[0], da[-1], db[0], db[-1]]),[0,3])
        ti, tf = ti-N*delta_t, tf+N*delta_t

        new_da = np.arange(ti, tf, delta_t)
        new_A = np.zeros_like(new_da)

        ind0 = np.where(new_da>=da[0])[0][0]
        ind1 = np.where(new_da>=da[-1])[0][0]
        x0 = new_da[ind0:ind1]

        f = interpolate.interp1d(da, A)
        f_p = f(x0)
        new_A[ind0:ind1] = f_p

        new_db = new_da
        new_B = np.zeros_like(new_da)
        ind2 = np.where(new_da>=db[0])[0][0]
        ind3 = np.where(new_da>=db[-1])[0][0]
        x1 = new_da[ind2:ind3]

        f = interpolate.interp1d(db, B)
        f_p = f(x1)

        new_B[ind2:ind3] = f_p


        return bs.time_domain(new_da, new_A), bs.time_domain(new_db, new_B)
    





      
def genrlzd_SNR1(wf_td,templ_gen, sc_ps,*templ_args,
                 N=None, cut_sc=None,taufix=None,
                 phase_opt=None, cut_pfind=None):    
    '''
    This is a generalized version of the SNR caluclator, built in 
    the matched filtering object as a method.
    
    Input:
        wf_td: time_domain object, real data from NR simualtions or
               from the GW detectors.
               
        templ_gen: function, this has to be a function that generates 
                   time_domain object. e.g.
                   
                   f = lambda f,d,t0,dt :monochr_templ(f,d,t0,dt, 
                                           amp=0.5*np.max(wf2.data))
                   
        sc_ps: Sensitivity_curve object
        
        *templ_args: ND arrays, can be scalar, 1D, 2D, ...
                     are those parameters on which the template generator depends
                     
        N: int, number of points to extend(zeropad) on the array domains
        
        cut_sc: tuple, (finit, fend) window where the sensitivity curve is 
                considered. outside of those limits it is zero padded.
        
        taufix: float, fixed timeshift at which the algorithm takes the SNR value
        
        phase_opt: bool, 
                   True-> computes everything based on the magnitude of the complex
                          SNR time series
                   False-> computes everything based on the real part of the complex
                           SNR time series
        cut_pfind: float, parameter that tells the algorithm to find the maximum ignoring
                   every timeshift value behind  it. WARNING! try not mixing it with taufix.
                   

    
    Output:
        SNR_rec_max: ND array, maximum recovered SNR calulated from the waveform, sensitivity
                     curve, and template by using the matched filtering algorithm.
    '''
    if phase_opt is None:
        phase_opt = True
    
    def SNR(*templ_args):
        tmpl_td = templ_gen(*templ_args, wf_td.dt)
        complex_SNR = matched_filtering(wf_td, tmpl_td , sc_ps)
        complex_SNR = complex_SNR.recovered_SNR3(N=N, cut_sc = cut_sc)
        
        if cut_pfind is None:
            pass
        else:
            complex_SNR = complex_SNR.cut((cut_pfind, complex_SNR.td[-1]))
        
        if taufix is None and phase_opt==True:
            SNR_rec_max = np.max(np.abs(complex_SNR.data))
        elif taufix is None and phase_opt==False:
            SNR_rec_max = np.max(complex_SNR.data.real)
        elif taufix is not None and phase_opt==True:
            ind = np.argwhere(complex_SNR.td>=taufix).flatten()[0]
            SNR_rec_max = np.abs(complex_SNR.data)[ind]
        elif taufix is not None and phase_opt==False:
            ind = np.argwhere(complex_SNR.td>=taufix).flatten()[0]
            SNR_rec_max = (complex_SNR.data.real)[ind]            
        
        if phase_opt==True:
            assert SNR_rec_max>0, 'warning! you are not actually using phase optimization'
        
        return SNR_rec_max
    
    SNR_vfunc = np.vectorize(SNR)

    return np.array(SNR_vfunc(*templ_args))



def genrlzd_SNR2(wf_td,templ_gen, sc_ps,max_templ_dur,*templ_args, taufix=None, phase_opt=True):    
    '''
    This is a generalized version of the SNR caluclator, built in 
    the matched filtering object as a method.
    
    Input:
        wf_td: time_domain object, real data from NR simualtions or
               from the GW detectors.
               
        templ_gen: function, this has to be a function that generates 
                   time_domain object. e.g.
                   
                   f = lambda f,d,t0,dt :monochr_templ(f,d,t0,dt, amp=0.5*np.max(wf2.data))
                   
        sc_ps: Sensitivity_curve object
        
        max_templ_dur: float, it has to be a multiple of the duration of the 
                       waveform or the template, which is needed to fix the padding
                       when doing the matched filtering algorithm.
        *templ_args: ND arrays, can be scalar, 1D, 2D, ...
                     are those parameters on which the template generator will depend
    
    Output:
        SNR_rec_max: ND array, maximum recovered SNR calulated from the waveform, sensitivity
                     curve, and template by using the matched filtering algorithm.
    '''

    def SNR(*templ_args):
        crop0, crop1 = [10, 3000], None
        tmpl_td = templ_gen(*templ_args, wf_td.dt,t_0=wf_td.t_0, amp= 1)
        complex_SNR = (matched_filtering(wf_td, tmpl_td , sc_ps).recovered_SNR1(max_templ_dur=max_templ_dur))[0]
        #complex_SNR = complex_SNR.cut((0, complex_SNR.td[-1]))
        return complex_SNR

    
    SNR_vfunc = np.vectorize(SNR)

    return np.array(SNR_vfunc(*templ_args))

def SNR_2D(wf, tp_func, var1, var2, sc, max_templ_dur=None):
    results = np.zeros_like(np.outer(var1, var2))
    for i, param0 in enumerate(var1):
        for j, param1 in enumerate(var2):
            if 1/param0 < param1 and 1/(8*param1) > wf.dt:
                template = tp_func(param0, param1, wf.dt)
                #wf_pm = ut.find_postm(wf)
                Mf0 = matched_filtering(wf.hp_get_tdobj(), template , sc).recovered_SNR1(max_templ_dur=max_templ_dur)
                ind = np.nanargmax(np.abs(Mf0[0].data))
                overlap = Mf0[0].data[ind]
                tau = Mf0[0].td[ind]
                phase = Mf0[1].data[ind]
                results[i,j] = overlap
            else:
                pass
    return results.T

            
    



import matplotlib.pyplot as plt
import matplotlib.image as mpimg
import numpy as np
import os
import lal

import utils0 as ut
import viz as viz
import snr as snr
from lal_wf import *
from core_wf import *
from eosdb import catalog


from scipy.integrate import simpson
from tqdm import tqdm

Y_22 = lal.SpinWeightedSphericalHarmonic(0,0,-2,2,2)

def catalogue_analysis(waveform_bank, func, sc,**func_kwargs):
    
    mode = func_kwargs['mode']
    
    R={}
    for i in range(0,len(waveform_bank['sims'])):
        print(waveform_bank['sims'][i])
        if mode=='core':
            wf = CoRe_NR_wf22(waveform_bank['db_path'], waveform_bank['sims'][i], waveform_bank['ex_param'], resol='best')
        elif mode=='lal':
            wf = lal_NR_wf1(waveform_bank['sims'][i],2,2, waveform_bank['ex_param']['dt'], waveform_bank['ex_param'])

        R[waveform_bank['sims'][i]] = func(wf, sc,**func_kwargs)
    return R



def full_analysis(waveform0, sc0, fbounds=None,Nf=None,
                  Nd=None, taufix=None,phase_opt=True,
                  mode=None, energetics=True, cut=None):
    
    if fbounds is None:
        fdounds = (500, 3500)
    if Nf is None:
        Nf=2
    if Nd is None:
        Nd = 2
    if mode is None:
        mode='core'
    
    tau1 = taufix
    pfind = cut
    
    wf_p0 = ut.find_postm(waveform0, before=0)
    templ = lambda f,d,dt,t_0=wf_p0.t_0,amp=1:ut.monochr_templ(f, d, dt, t_0,amp, phase=0)
    
    # Matched filtering
    f = np.linspace(fbounds[0],fbounds[1], Nf)
    d = np.linspace(0.1e-3, 1*wf_p0.dur, Nd) 
    
    f_mesh0, d_mesh0 = np.meshgrid(f,d)
    
    S = snr.genrlzd_SNR1(wf_p0.hp_get_tdobj(),templ,sc0,
                   f_mesh0, d_mesh0, N=2000, cut_sc=None,
                   taufix=tau1, phase_opt=True, cut_pfind=pfind)
    res = ut.optimal_template(wf_p0, sc0, S,
                 f_mesh0, d_mesh0, 3, tau1,
                 pfind, 2000, cut_sc=None,
                 plot=False, zoom=-5)
    fmax, dmax, taumax, phimax, smax, sopt = res
  
    
    if energetics==True:
        dist_m = wf_p0.mdta['dist_mpc'] * 3.086e22
        h22 = (1/Y_22)*(wf_p0.hp_get_tdobj().data+1j*wf_p0.hc_get_tdobj().data)
        _, L = ut.energylm(2,2, h22, wf_p0.td, dist_m)
        _, T = ut.angmom_Z_lm(2,2, h22, wf_p0.td,dist_m)
        e_rad, J_rad = simpson(L.data, L.td), simpson(T.data, T.td)
    
    result = {}

    result['best_f'] = (fmax, 'Hz')
    result['best_d'] = (dmax, 's')
    result['best_tau'] = (taumax, 's')
    result['best_phi'] = (phimax, 'rad')
    
    result['wf_d'] = (wf_p0.dur, 's')
    result['wf_d_wei'] = (wf_p0.amp_wei_time(),'s')
    result['wf_f_wei'] = (wf_p0.amp_wei_freq(), 'Hz')
    
    result['SNR_rec'] = (smax, 'dmless')
    result['SNR_opt'] = (sopt, 'dmless')

    
    result['mdta'] = dict(wf_p0.mdta)
    
    result['mdta']['E'] = (e_rad,'Msun')
    result['mdta']['Jz'] = (J_rad, 'Msun^2')
    
    if result['mdta']['ecc']=='':
        result['mdta']['ecc']=np.nan
    
    if mode=='core':
        if result['mdta']['eos'] in EOS_masses().keys():
            result['mdta']['maxm_TOV'] = EOS_masses()[result['mdta']['eos']]['maxm_TOV']
        else:
            result['mdta']['maxm_TOV'] = np.nan
            
    elif mode=='lal':
        cat = catalog.load_catalog()
        eos = result['mdta']['eos'].split('hybrid-pp-')[::-1][0] 
        if eos.upper()+'_LAL'  in cat.keys():
            result['mdta']['maxm_TOV'] = cat[eos.upper()+'_LAL'].tov.mg_max
        elif eos.upper()+'_EPP' in cat.keys():
            result['mdta']['maxm_TOV'] = cat[eos.upper()+'_EPP'].tov.mg_max
        else:
            result['mdta']['maxm_TOV'] = np.nan
    
    return result

#def wf_vizlysis(wf, templ_bank, path,taufix=None, sc_name='et_d',mode='core'):
#    path_code = os.path.expanduser('~/REPOS/sebas-training/thesis/Postmerger/')
#    sc = ut.ASD(sc_name, path_code)
#    templ_func = lambda fr,du,dt,t_0=0, amp=1 :ut.monochr_templ(fr,du,dt,t_0,amp)
#    freq_mesh, dur_mesh = np.meshgrid(templ_bank['frequency'], #templ_bank['duration'])
#    
#    fig = plt.figure(figsize=(20,20))
#    
#    if mode=='lal':
#        direc = path+'/'+(wf.mdta['name'].split(':R')[0])
#        if os.path.exists(direc):
#            shutil.rmtree(direc)
#        os.mkdir(direc)
#        new_path = path+'/'+(wf.mdta['name'].split(':R'))[0]
#
#        wf_p = ut.find_postm(wf).hp_get_tdobj()    
#        N = np.round(np.max(templ_bank['duration'])/wf_p.dur, decimals=0)
#        tmax=N*wf_p.dur
#
#        mask = viz.rjct_criteria_monochtempl(freq_mesh, dur_mesh, wf_p.dt)
#        S = genrlzd_SNR(wf_p,templ_func,sc,tmax,freq_mesh, dur_mesh,taufix=taufix)
#
#        fig.clear()
#        fig, ind = viz.best_SNR(S,freq_mesh, dur_mesh, 
#                          rjct_criteria=mask, fig=fig,
#                          xlabel='frequency[Hz]',ylabel='duration[s]')
#
#        fig.savefig(new_path+'/'+'surf.jpg')
#
#        fig.clear()
#        wf.show_wf(mdta=True, xlabel='time[s]', ylabel='strain', fig=fig)
#        fig.savefig(new_path+'/'+'wf.jpg')
#
#        fig.clear()
#        viz.best_monochr(ind,freq_mesh, dur_mesh, wf_p, sc, fig=fig)
#        fig.savefig(new_path+'/'+'best_match.jpg')
#        plt.close()
#    elif mode=='core':
#        direc = path+'/'+(wf.mdta['name'].split(':R')[0])
#        if os.path.exists(direc):
#            shutil.rmtree(direc)
#        os.mkdir(direc)
#        new_path = path+'/'+(wf.mdta['name'].split(':R'))[0]

#        wf_p = ut.find_postm(wf).hp_get_tdobj()    
#        N = np.round(np.max(templ_bank['duration'])/wf_p.dur, decimals=0)
#        tmax=N*wf_p.dur
#
#        mask = viz.rjct_criteria_monochtempl(freq_mesh, dur_mesh, wf_p.dt)
#        S = genrlzd_SNR(wf_p,templ_func,sc,tmax,freq_mesh, dur_mesh)
#
#        fig.clear()
#        fig, ind = viz.best_SNR(S,freq_mesh, dur_mesh, 
#                          rjct_criteria=None, fig=fig,
#                          xlabel='frequency[Hz]',ylabel='duration[s]')
#
#        fig.savefig(new_path+'/'+'surf.jpg')
#
#        fig.clear()
#        wf.show_wf(mdta=True, xlabel='time[s]', ylabel='strain', fig=fig)
#        fig.savefig(new_path+'/'+'wf.jpg')
#
#        fig.clear()
#        viz.best_monochr(ind,freq_mesh, dur_mesh, wf_p, sc, fig=fig)
#        fig.savefig(new_path+'/'+'best_match.jpg')
#        plt.close()
#    return print('the analysis of the database has been stored in '+'\n'+new_path)

#path_to = os.path.expanduser('~/analysis') #path to store plots
#snr.wf_vizlysis(wf0, template_bank, path_to)
#path_to_img =  path_to +'/'+ lal_NR_wf(lal_list[0], deltat, src).mdta['name']
#path_to_img =  path_to +'/'+ waveform_bank['core'][0]
#viz.show_results(path_to_img)



def show_results(path_to_img):
    img = [mpimg.imread(path_to_img+'/'+'surf.jpg'),
           mpimg.imread(path_to_img+'/'+'wf.jpg'),
           mpimg.imread(path_to_img+'/'+'best_match.jpg')]

    size=(20,10) 

    ax0 = plt.figure(figsize=size).add_subplot()
    ax1 = plt.figure(figsize=size).add_subplot()
    ax2 = plt.figure(figsize=(30,20)).add_subplot()

    ax0.axis('off')
    ax1.axis('off')
    ax2.axis('off')

    ax0.imshow(img[0])
    ax2.imshow(img[2])
    ax1.imshow(img[1])

    ax0.set_ylim(1000, 200)
    ax0.set_xlim(130, 1300)

    ax1.set_ylim(1000, 450)
    ax1.set_xlim(130, 1300)

    ax2.set_ylim(1300, 160)
    ax2.set_xlim(130, 1300)
    plt.show()
    
    
def catalog_plot(q, R, ax=None, mode='lal',show_ticks=True, EOS_color_patellete=None, step=4):
    
    list_ticks = list(R.keys())
    ind = (np.arange(len(q)+1, step=step))
    ind[1:] = np.arange(len(q)+1, step=step)[1:]-1
    sims1 = np.array(list_ticks)[ind]
    
    C = EOS_color_patellete
    
    if mode=='lal':
        rel = lal_relabeling(R)
        for i, sim in enumerate(sims1):
            #name = R[sim]['mdta']['name']
            #sims1[i] = rel[name]
            sims1[i] = rel[sim]

    if ax is None:
        ax = plt.axes()
    if show_ticks==False:
        ax.set_xticks(ind, sims1, rotation= 'vertical')
        ax.tick_params(axis='x', which='both', bottom=False, top=False, labelbottom=False)
    else:
        ax.set_xticks(ind, sims1, rotation= 'vertical')
    E = []
    for i, q in enumerate(q):
        eos = R[list_ticks[i]]['mdta']['eos']
        if eos not in E:
            E.append(eos)
            ax.scatter(i, q, c= C[eos], s=2, label=eos)
        else:
            ax.scatter(i, q, c= C[eos], s=2)
    ax.margins(x=0.009, y=0.5)
    ax.grid()
    ax.locator_params(axis='y', nbins=6)
    return ax

def which_sim(D, mask1, mask2):
    m = np.logical_and(mask1, mask2)    
    L = np.array(list(D.keys()))
    
    return L[m]


def lal_relabeling(A):
    j=1
    D ={}
    for i, name in enumerate(A.keys()):
        eos = A[name]['mdta']['eos']
        if eos=='hybrid-pp-APR4':
            if j<=10:
                #D[A[name]['mdta']['name']] = f'LAL:000{j}'
                D[name] = f'LAL:000{j}'
            else:
                #D[A[name]['mdta']['name']] = f'LAL:00{j}'
                D[name] = f'LAL:00{j}'
            j+=1


    for i, name in enumerate(A.keys()):
        eos = A[name]['mdta']['eos']
        if eos=='SHT':
            if j<=10:
                #D[A[name]['mdta']['name']] = f'LAL:000{j}'
                D[name] = f'LAL:000{j}'

            else:
                #D[A[name]['mdta']['name']] = f'LAL:00{j}'
                D[name] = f'LAL:00{j}'

            j+=1

    for i, name in enumerate(A.keys()):
        eos = A[name]['mdta']['eos']
        if eos=='hybrid-pp-H4':
            if j<=10:
                #D[A[name]['mdta']['name']] = f'LAL:000{j}'
                D[name] = f'LAL:000{j}'

            else:
                #D[A[name]['mdta']['name']] = f'LAL:00{j}'
                D[name] = f'LAL:00{j}'

            j+=1

    for i, name in enumerate(A.keys()):
        eos = A[name]['mdta']['eos']
        if eos=='hybrid-pp-H4':
            if j<=10:
                #D[A[name]['mdta']['name']] = f'LAL:000{j}'
                D[name] = f'LAL:000{j}'

            else:
                #D[A[name]['mdta']['name']] = f'LAL:00{j}'
                D[name] = f'LAL:00{j}'
            j+=1

    for i, name in enumerate(A.keys()):
        eos = A[name]['mdta']['eos']
        if eos=='hybrid-pp-MS1':
            if j<=10:
                #D[A[name]['mdta']['name']] = f'LAL:000{j}'
                D[name] = f'LAL:000{j}'

            else:
                #D[A[name]['mdta']['name']] = f'LAL:00{j}'
                D[name] = f'LAL:00{j}'

            j+=1

    for i, name in enumerate(A.keys()):
        eos = A[name]['mdta']['eos']
        if eos=='hybrid-pp-SLy':
            if j<=10:
                #D[A[name]['mdta']['name']] = f'LAL:000{j}'
                D[name] = f'LAL:000{j}'
            else:
                #D[A[name]['mdta']['name']] = f'LAL:00{j}'
                D[name] = f'LAL:00{j}'

            j+=1
    return D



import numpy as np
import base as bs
import utils_wf as utw 
from waverep import save_gw_ligo as gwlvc
#import lalsimulation as ls
import lal

def get_lalwfattr(wf_gwlvc):
    '''
    Helper function for getting the metadata from 
    the lal waveforms
    '''
    m2, m1 = np.sort(np.array([wf_gwlvc.grav_mass1,
              wf_gwlvc.grav_mass2]))
    spin_1, spin_2 = [np.linalg.norm(np.array(wf_gwlvc.spin1)),
                    np.linalg.norm(np.array(wf_gwlvc.spin2))]
    
    ecc = wf_gwlvc.eccentricity
    if ecc=='':
        ecc=np.nan
    
    attr ={'name': wf_gwlvc.name, 
           'eos':wf_gwlvc.eos_name,
           'grav_mass1':m1,
           'grav_mass2':m2,
           'lambda_tilde': np.nan,
           'Xeff': utw.chi_eff(m1,m2, spin_1,spin_2),
           'spin1':spin_1,
           'spin2':spin_2,
           'ecc':ecc
           }
    return attr 


def lal_NR_wf1(path, l, m, dt, src_mdata):
        
    '''
    This is a function that creates a waveform object from lal
    
    Input:
         path: str,
                  path to the folder in which the database is located
         dt : float,
              sampling rate of the waveform usually one should pick
              something like 1/40000
               
         src_mdta: dictionary,
                   contains information about the sky
                   positioning of the source:(typical params)
                       *src_mdta['inclination']
                       *src_mdta['dist_mpc']
                  
    Output: 
         waveform object that holds hp, hc, time array and metadata 
               
    '''
    
    MSUN_METER = lal.MSUN_SI * lal.G_SI/(lal.C_SI**2)
    MSUN_SECOND = MSUN_METER / lal.C_SI
    
    inclination = src_mdata['inclination']
    dist_mpc = src_mdata['dist_mpc']
    distance = dist_mpc * lal.PC_SI * 1.0e6
    
    wf_file = path
    wfrm = gwlvc.WaveFormFile(wf_file)
    
    ct,ch = wfrm.strain_units_msol(l,m, samp_rate=MSUN_SECOND/dt)
    
    fac = lal.SpinWeightedSphericalHarmonic(inclination,0,-2,l,m)

    hp = fac * ch.real * MSUN_METER / distance
    hc = fac * ch.imag * MSUN_METER / distance
    
    t =  ct * MSUN_SECOND
    
    wf_data = get_lalwfattr(wfrm)
    src_mdata.update(wf_data)
    
    return bs.waveform(t, hp, hc, src_mdata)



import numpy as np
import matplotlib.pyplot as plt
from scipy import interpolate
from scipy import signal
import h5py

import mfNRcatpy.snr as snr

class power_spectrum(object):

    def __init__(self,frequencies, PSD):
        self.data = PSD
        self.fd = frequencies
        self.df = frequencies[1] - frequencies[0]
        
            
    def PSD_flip_mirror(self, F_in):
        '''
        This method takes a PSD that is only
        defined in the positive side of the time/frequency
        space.

        It symmetrizes the array by mirroring its complex
        conjugate to the left

        Used attributes:
            self.data --> 1D array, data in the frequency domain, must be equally sampled

        Input:
            F_in --> 1D array, frequency range in which one wants to extend it

                     Remark :
                     F_in should be growing ordered

        Output:
            power_spectrum object:   new extended(to the left) array, which is equally sampled
                                     by using its own complex conjugate on the left half plane 
                                     in the frequency domain
                                     
                                     power_spectrum object-> which hold the following information:
                                     self.fd-> equally spaced array
                                     self.data-> equally spaced array
        '''


        PSD_new = np.zeros_like(F_in)
        msk0 = np.logical_and(F_in>=0, F_in<=np.max(F_in)) 
        msk1 = np.logical_and(F_in>=-np.max(F_in), F_in<0) 
        PSD_new[msk0] = self.data
        PSD_new[msk1] = (self.data[1:].conjugate())[::-1]

        return power_spectrum(F_in, PSD_new)
            
        
        
class time_domain(object):
    
    def __init__(self, t, arr):
        self.dt = t[1] - t[0]
        self.dur = t[-1] - t[0]
        self.t_0 = t[0]
        self.td = t
        self.data = arr
        
    def to_fd(self):
        '''
        This is a method that takes a time_domain objects
        and tranforms it to its frequency domain form, by using
        numpy fft algorithm.
        
        Used attributes:
            self.data: 1D array
            delf.dt: float, sampling rate of the time doain data
            
        Input: None
            
        output:
            frequecy_domain object-> holds the following attributes:
                df: sampling rate in the frequency domain
                fd: frequecy array 
                data: data array       
                
        '''
        frequencies = np.fft.fftfreq(self.data.size, self.dt)
        data_fd = np.fft.fft(self.data)*self.dt
        
        frequencies = np.fft.fftshift(frequencies)
        data_fd = np.fft.fftshift(data_fd)
        
        return freq_domain(frequencies, data_fd)
    
    def spectral_phase(self, thresh=None):
        
        '''
        This function computes the spectral phase of any 
        time series. First cleaning all the noise produced 
        when computing the FFTs, using a threshold based 
        technique.
        
        tresh: float, (0, 0.7]-> recommended
        '''
        
        if thresh is None:
            thresh = 0.1
            
        y_f = self.to_fd().data.copy()
        freqs = self.to_fd().fd.copy()
        
        y_f[y_f<thresh*np.abs(y_f).max()] = 0
        
        phase = np.arctan2(y_f.imag, y_f.real)
        
        return freq_domain(freqs, phase)
        
    
    def cut(self, t_cut):
        '''
        This method crops the time_domain objects, to a region of interest
        
        Used attributes:
            self.td -> 1D array
            self.data -> 1D array
        
        Input:
            t_cut -> tuple, should be in miliseconds
                       t_cut[0]=t_lower
                       t_cut[1]=t_higher
        Output:
            time_domain object, which holds the following attributes
                dt, sampling rate
                dur, duration
                t_0, initial phase
                td, time array 
                data, data array
        '''
        msk = np.logical_and(self.td>=t_cut[0], self.td<=t_cut[1])

        new_td = self.td[msk]
        new_data = self.data[msk]
        
        return time_domain(new_td, new_data)
    
    def spectrogram_from_td(self, ax=None, axins=None):
        F = np.fft.fftfreq(self.td.size, self.dt)
        fs = F[1]-F[0]
        N = F.size
        length_pieces = N//10
        f, t, Sxx = signal.spectrogram(self.data, 
                                    fs=1/(self.dt),nperseg=length_pieces, 
                                    nfft=32*length_pieces,noverlap=N//11)

        T, F = np.meshgrid(t, f)
        ind = np.unravel_index(np.argmax(Sxx), Sxx.shape)
        Fmax = F[ind]
        
        T = (T+self.t_0)

        fig = plt.figure(dpi=900)
        
        if ax is None:
            ax = fig.add_subplot()
            
        ax.pcolormesh(T*1e3, F, np.sqrt(Sxx), 
                            shading='gouraud',cmap='cubehelix',
                           antialiased=True,rasterized=True)

        ax.set_ylabel('Frequency [Hz]')
        ax.set_xlabel('Time [ms]')
        
        if axins is not None:

            axins = ax.inset_axes([0.7, 0.7, 0.2, 0.2])
            axins.pcolormesh(T*1e3, F, np.sqrt(Sxx), 
                        shading='goraud', rasterized=True,
                        cmap='YlGnBu', antialiased=True)

            axins.set_xlim(T.min()*1e3, (T.min()+self.dur/5)*1e3)
            if self.dur >7e-3:
                axins.set_ylim(Fmax-500, Fmax+500)
            else:
                axins.set_ylim(10, 5000)
            axins.tick_params(left = False, right = False , labelleft = False ,
                            labelbottom = False, bottom = False)


            ax.indicate_inset_zoom(axins, edgecolor="black");

        if self.dur <=10e-3:
            ax.set_ylim(10, 40000)
        else:
            ax.set_ylim(10, 4000)        
        return ax

        
        
        
class freq_domain(object):
    '''
    Input:
            fft ordered arrays
    '''
    def __init__(self, frequencies, arr_fd):
        self.df = frequencies[1]-frequencies[0]
        self.fd =frequencies
        self.data = arr_fd
        
    def to_td(self):
        time = np.ones_like(self.data)
        data_td = np.fft.ifft(self.data)*self.df
        return time_domain(time, data_td)

            
    
    
class waveform(object):   
    
    def __init__(self, t, hp, hc, src_mdta):
        self.t_0 = t[0]
        self.td = t
        self.dt = self.td[1]-self.td[0]
        self.dur = self.td[-1]-self.td[0]
        self.hp, self.hc = hp, hc
        self.GW_ampl = np.abs(self.hp+1j*self.hc)
        self.GW_phase = np.unwrap(-np.angle(self.hp+1j*self.hc))
        self.GW_inst_freq = np.gradient(self.GW_phase,self.dt,edge_order=2)*(1/(2*np.pi))
        self.mdta = src_mdta
        self.GW_tmax = self.td[np.argmax(self.GW_ampl)]
        self.td = self.td-self.GW_tmax
        self.GW_tmax = 0.0
        
    def sopt(self, sc, N=None):
        if N is None:
            N=5000
        ones = np.ones_like(self.td)
        ones = time_domain(ones, ones)
        
        sopt = snr.matched_filtering(self.hp_get_tdobj(), ones, sc).optimal_SNR(N=N)
        
        return np.abs(sopt.data).max()
        
    #def GW_tmax(self):
    #    ind = np.argmax(self.GW_ampl)
    #    return self.td[ind]
        
    def amp_wei_time(self):
        return 2 * np.sum(self.GW_ampl*(self.td-self.td[0]))/np.sum(self.GW_ampl)   
        
    def amp_wei_freq(self):
        return np.abs(np.sum(self.GW_ampl*self.GW_inst_freq)/np.sum(self.GW_ampl))
        
        
    def show_wf(self, mdta=False, fig=None, **kwargs):
        '''
        this method provides a plotting functionality 
        which can show metadata alonside with the wave itself
        
        Used attributes:
            self.td
            self.hp
            self.hc
            self.mdta
        
        Input:
            mdta: bool, default->False, when True the metadata will
                  occupy some space in the plot.
            fig: pyplot figure object, default->None,
                 One can pass a custom fig created beforehand, e.g.
                     fig = plt.figure(figsize=(15,5))
                     
            **kwargs: keyword args for the plotting:
                      xlabel, ylabel, xlim, ylim, ...
        
        Output:
            fig
        '''
        
        D = self.get_mdta()
        #dist = D['dist_mpc']
        if  len(D['name'])>15:
            D['name'] = D['name'][:10] +'...'
            
        table = np.array([list(D.keys()), list(D.values())]).T
        
        ind_tmerg = np.argmax(self.GW_ampl)
        
        
        if fig is None:
            fig = plt.figure(figsize=(7, 4.5), dpi=400)
            fig.subplots_adjust(wspace=0.5, hspace= 0.5)
        if mdta is True:
            gs = fig.add_gridspec(4,20)
            ax0 = fig.add_subplot(gs[:,:10])
            #ax0.set_ylabel(f'strain at {dist}Mpc ')
            ax0.set_xlabel('time [ms]')
            ax1 = fig.add_subplot(gs[:,10:])
            
            ax0.plot(self.td*1e3, self.hp, label=r'$h_+$', c='green')
            ax0.plot(self.td*1e3,self.hc, label=r'$h_x$', c='black')
            ax0.legend(loc='upper right')
            ax1.set_title('source parameters')
            table = ax1.table(table, loc = 'center', cellLoc='center')
            table.auto_set_font_size(False)
            table.set_fontsize(8.5)
            ax1.axis('off')
        else:
            #name, eos = D['name'], D['eos']
            ax0 = fig.add_subplot()
            #ax0.set_title(f'{name}, EOS:{eos}')
            #ax0.set_ylabel(f'strain at {dist}Mpc ')
            ax0.set_xlabel('time [ms]')
            ax0.plot(self.td*1e3, self.hp, label=r'$h_+$', c='green')
            ax0.plot(self.td*1e3,self.hc, label=r'$h_x$', c='black')
            ax0.axvline(self.GW_tmax*1e3, color='red',
                       linestyle='dashed')#, label=r'$\mathrm{Max(h_{22})}$')
            ax0.legend(loc='upper right')
            
        return ax0

            
    def h_AP(self, theta1, theta2):
        '''
        Antenna pattern using the coeficients
        depending on the angles which determine the orientation of the source
        '''
        wave = time_domain(self.td, np.ones_like(self.td))
        return wave
    
    def get_mdta(self):
        '''
        This method is used to give a nice formatted
        printed version of the metadata
        
        Used attributes:
            self.mdta: dictionary
            
        Input: None
        
        Output:
            show: str, formatted table
        '''
      
        keys = list(self.mdta.keys())
        values = list(self.mdta.values())
        A = {}
        for i in keys:
            A.update({i:self.mdta[i]})
        return A
    
    def hp_get_tdobj(self):
        '''
        This method is used to create time_domain objects
        from just one of the polarizations, hp
        
        Used attributes:
            self.hp: 1D array
            self.td: 1d array
            
        Input: None
        
        Output:
            time_domain object, which holds the following attributes
                dt, sampling rate
                dur, duration
                t_0, initial phase
                td, time array 
                data, data array
        '''
        wave = time_domain(self.td, self.hp)
        return wave
    
    def hc_get_tdobj(self):
        '''
        This method is used to create time_domain objects
        from just one of the polarizations, hc
        
        Used attributes:
            self.hc: 1D array
            self.td: 1d array
            
        Input: None
        
        Output:
            time_domain object, which holds the following attributes
                dt, sampling rate
                dur, duration
                t_0, initial phase
                td, time array 
                data, data array
        '''        
        wave = time_domain(self.td, self.hc)        
        return wave
        
    

class Sensitivity_curve(object):
    '''
    Input:
        non-equally sampled & growing ordered arrays
        [a1, a2, aj] with aj>...>a2>a1
        
        frequencies: 1D array, frequency bins
        ASD : 1D array, amplitude spectral density 
        name: str, this is used for plotting purposes later on           
            
    '''
    def __init__(self, frequencies, ASD, name):
                      
        self.fd = frequencies
        self.ASD = ASD
        self.PSD = ASD**2
        self.name = name
                              
    def cut(self, bounds):
        '''
        This method allow the matched filtering algorithm
        to just take into account certain region of the 
        frequency spectrum, to make the analysis. later on
        the other neglected regions are zeropadded, to ignore 
        weird behavior when applying fft to sharp regions on 
        some datasets
        
        Used attributes:
            self.fd: 1D array, non-equally sampled from ligo/virgo/kagra files
            sefl.ASD: 1D array, non-equally sampled from ligo/virgo/kagra files
            
        Input:
            bounds, tuple-> (f_lower,f_higher)
        
        Output:
            sensitivity_curve object, which holds the following attributes:
                self.df
                self.fd-> 1D array, non equally sampled
                self.ASD-> 1D array, non equally sampled
                self.PSD-> 1D array, non equally sampled
        '''
        msk = np.logical_and(self.fd>=bounds[0],self.fd<=bounds[1])
        new_fd = self.fd[msk]
        new_ASD = self.ASD[msk]
        
        return Sensitivity_curve(new_fd, new_ASD, self.name)
    
    def PSD_interp(self, F_in):
        '''
        This method is used to take the ligo/virgo/...
        design curves to be sampled in the same way as
        the templates, and waveforms.
        
        Used attributes:
            self.fd-> 1D array, non equally sampled
            self.ASD-> 1D array, non equally sampled
            
        Input:
            F_in, 1D array. is also growing ordered
                                a1, a2, aj 
                                with aj>...>a2>a1        
        Output:
            power_spectrum object-> which hold the following information:
                                    self.fd-> equally spaced array
                                    self.data-> equally spaced array
        
        '''
        
        msk = np.logical_and(F_in>=self.fd[0], F_in<=np.max(self.fd))
        new_F = F_in[msk]

        func = interpolate.interp1d(self.fd, self.ASD)
        new_ASD = func(new_F)
        
        return power_spectrum(new_F, new_ASD**2)   

    
    def inv_psd_extension(self,F_in):
        '''
        This method completes the spectrum for the design curves
        because usually they come defined just in the right frequency 
        plane
        
        Input:
            F_in, 1D array-> is growing ordered
                                a1, a2, aj 
                                with aj>...>a2>a1 
        
                                F_in must have a constant
                                sampling rate
        
        Output:
            power_spectrum object-> which holds the following information:
                        self.fd-> equally spaced array
                        self.data-> equally spaced array
        
        '''
        # 1. PSD needs to be equally spaced as F_in
        #    to be extended towards those regions in
        #    which it is not defined
        
        interp_data_PSD = (self.PSD_interp(F_in)).data  # this is a powerspectrum object
                                                        # we use .data to extract the PSD
        
        inv_PSD0 = 1/(interp_data_PSD)
        
        # crop the positive part of the target frequency space
        # crop the positive part of frequency space
        msk1 = np.logical_and(F_in>=0, F_in<=np.max(F_in))
        F_sc1 = F_in[msk1]

        # zero pad the inverse PSD on those regions in which it is not defined

        inv_PSD1 = np.zeros_like(F_sc1)
        msk2 = np.logical_and(F_sc1>=self.fd[0], F_sc1<=np.max(self.fd))
        inv_PSD1[msk2] = inv_PSD0
        
        PSD_extended = inv_PSD1

        return power_spectrum(F_sc1, PSD_extended)   
    



import numpy as np
from watpy.coredb.coredb import *
import lal
import mfNRcatpy.base as bs
import mfNRcatpy.utils as ut
import mfNRcatpy.utils_wf as utw 
from IPython.utils import io



def get_CoRewfattr(wf_CoRedb):
    '''
    Helper function for getting the metadata from 
    the CoRe waveforms
    '''
    spin1 =np.array(wf_CoRedb.md.data['id_spin_starA'].split(',')).astype('float')
    spin2 =np.array(wf_CoRedb.md.data['id_spin_starB'].split(',')).astype('float')
    
    spin_norm1, spin_norm2 = np.linalg.norm(spin1),np.linalg.norm(spin2)
    spin_z1, spin_z2 = spin1[2],spin2[2]
    
    m2, m1 = np.sort(np.array([np.float64(wf_CoRedb.md.data['id_rest_mass_starA']),
             np.float64(wf_CoRedb.md.data['id_rest_mass_starB'])]))
    

    attr ={'name':wf_CoRedb.md.data['database_key'],
           'eos':wf_CoRedb.md.data['id_eos'],
           'grav_mass1':m1,
           'grav_mass2':m2,
           'lambda_tilde': np.float64(wf_CoRedb.md.data['id_Lambda']),
           'Xeff': utw.chi_eff(m1,m2, spin_norm1,spin_norm2),
           'spin1':spin_norm1,
           'spin2':spin_norm2,
           'ecc': wf_CoRedb.md.data['id_eccentricity']
          }
    
    if attr['ecc'] is None or attr['ecc']=='':
        attr['ecc']=0.0
    else:
        attr['ecc']=np.float64(attr['ecc'])
    
    return attr

def resolution_selector(cdb, sim,resol):
    if resol is None:
        f = [[res,cdb.sim[sim].run[res].md.data['grid_spacing_min']] for res in cdb.sim[sim].run.keys()]
        f = np.array(f)
        run = input(f'simulation {sim} has several available resolutions(grid_spacing_min) \n {f} \n where the smaller the better, please choose one \n')
        min_gridres = cdb.sim[sim].run[run].md.data['grid_spacing_min']
        
    elif resol=='best':
        f = [[res,cdb.sim[sim].run[res].md.data['grid_spacing_min']] for res in cdb.sim[sim].run.keys()]
        f = np.array(f)
        ind = np.argmin(f[:,1].astype('float'))
        
        run = f[ind, 0]
        min_gridres = cdb.sim[sim].run[run].md.data['grid_spacing_min']

    else:
        run = resol
        min_gridres = cdb.sim[sim].run[run].md.data['grid_spacing_min']
    
    return run, min_gridres 



def CoRe_NR_wf22(db_path, sim, src_mdta,resol='best',sync_db=False, energetics=False):
    
    '''
    This is a function that creates a waveform object from the 
    CoRe colaboration catalogue
    
    This waveforms come as r*h22 as a function of a timelike coordinate U/M, 
    where instead of taking M is the total mass of the system
    in solar masses.
    
    Where [r]=[Msun], h22 is dimensionless, [U/M]=[Msun/Msun]
    
    
    Input:
         db_path: str,
                  path to the folder in which the database is located
         sim : str,
               name of the simulation, more info can be found in the 
               webpage: http://www.computational-relativity.org/gwdb/
               
         src_mdta: dictionary,
                   contains information about the sky
                   positioning of the source:(typical params)
                       *src_mdta['inclination'] in radians
                       *src_mdta['phiref'] in radians
                       *src_mdta['dist_mpc'] in megaparsec
                       
         sync_db: bool, 
                  set False by default. When True, updates the database
                  to the latest version from its original repository
                  
    Output: 
         waveform object that holds hp, hc, time array and metadata 
               
    '''
    with io.capture_output() as captured: ## avoiding nasty printing
        cdb = CoRe_db(db_path)
        if sync_db is True:
            idb = CoRe_index(db_path, lfs=False, verbose=False)
            idb.sync_database(lfs=True, verbose=False)
            
            
    run, min_gridres = resolution_selector(cdb, sim, resol)

    try:
        with io.capture_output() as captured:  
            wf = CoRe_db(db_path).sim[sim].run[run]
    except ValueError:
        print(f' there is a problem with simulation {sim} run {run}, min_grid_res={min_gridres}')
            
    i = src_mdta['inclination']

    Y_22 = lal.SpinWeightedSphericalHarmonic(i,0,-2,2,2)  

    rh22 = wf.data.read('rh_22')
    
    U_M, rh = rh22[:,0], (rh22[:,1] + 1j*rh22[:,2])*Y_22
    
    attr = {}
    attr = get_CoRewfattr(wf)
    keys = list(attr.keys())
    
    for i in range(len(keys)):
        src_mdta[keys[i]] = attr[keys[i]]
                
    m_tot = float(attr['grav_mass1'])+float(attr['grav_mass2'])
    dist_msun = (float(src_mdta['dist_mpc'])*3.0856776e+22) * (1/(1476.62*m_tot))
    
    m_tot_sec = m_tot * 4.925e-6
            
    return bs.waveform(U_M*m_tot_sec, rh.real/dist_msun, rh.imag/dist_msun, src_mdta)



from setuptools import setup, find_packages
from pathlib import Path
this_directory = Path(__file__).parent
ld = (this_directory / "README.md").read_text()
print(ld)

find_packages(where='mfNRcatpy')

setup(
    name='mfNRcatpy',
    version='0.0.12',
    long_description=ld,
    long_description_content_type='text/markdown',
    packages=find_packages() + find_packages(where='mfNRcatpy'),
    install_requires=['numpy',
                      'scipy',
                      'core-watpy',
                      'h5py',
                      'matplotlib',
                      'lalsuite',
                      'numba'],
    author = 'Sebastian Gomez Lopez',
    author_email= 'segomezlo@unal.edu.co',
    url = 'https://github.com/sebastiangomezlopez/mfNRcatpy'

    )



import numpy as np
from scipy import signal
import pathlib
import os
from scipy.integrate import cumtrapz, simpson
import matplotlib.pyplot as plt
from matplotlib.ticker import FormatStrFormatter

import mfNRcatpy.base as bs
import mfNRcatpy.snr as snr





def ASD(filename,directory):
    '''
    This function imports sensitivity design curves from 
    current detectors ligo, virgo, kagra, Einstein telescope
    
    Input:
    
        filename: str, can be chosen from the keywords:
                  ---> 'aligo', 'advirgo', 'et_d', 'kagra'
        directory: str, path to the folder in which the files 
                   are stored
     
    Output:    
        sensitivity curve object:
               --> which contains  non-equally sampled  ASD data
               as a function of the frequency
               --> this objects have some builtin methods like interpolation and extension
                   , that come in handy when doing a matched filtering algorithm.
     
     
     Remarks: PLEASE READ!
         1. directory must end in /
         2. detector ASDs must be downloaded from ligo webpage
            https://dcc.ligo.org/LIGO-T1500293/public, or some 
            similar webpage, with updated versions of the data.
    '''
    freq, strain = np.loadtxt(str(directory)+str(filename)+'.txt').T
    
    return bs.Sensitivity_curve(freq, strain, filename)


def monochr_templ(freq, duration, dt, t_0=0, amp=1, phase=0, alpha = None):
    '''
    This is a function that computes a harmonic function of time
    
    Input:        
        freq: float
        duration: float
        t_0: float, initial phase of the wave
        dt: float, sampling rate of the wave
        
        amp: float, default-> 1e-20, similar order of magnitude as 
             waveforms of GW's from binary systems
    
    Output:
        time_domain object, which holds the following attributes
            dt, sampling rate
            dur, duration
            t_0, initial phase
            td, time array 
            data, data array

    '''
    
    if alpha is None:
        alpha = 0
        
    time = np.arange(t_0, (t_0 + duration), dt)
    tukey  = signal.windows.tukey(time.size, alpha=alpha)
    
    templ_r = amp * np.sin((2*np.pi*freq*time) + phase) * tukey
    
    return bs.time_domain(time, templ_r)
   

def rjct_criteria_monochtempl(f_mesh,d_mesh,dt, wl=2):
    '''
    This funtion creates a mask wich rejects points in the 2D parameter space for 
    monochromatic wave templates. It does it based on:
        *Nyquist thorem, templates must have a sampling rate haigh enough to avoid 
                         aliasing
        *templates with less than one full cycle are rejected
    
    Input:
        f_mesh:2D array, frequency mesh
        d_mesh:2D array, duration mesh
        dt: float, sampling rate
    Output:
        msk1, mask(2D numpy array of bools) which is False at those rejected spots
              in the param space where one gets an unaccepted template
     
    Example:
        accepted_tmpl = rjct_criteria(templ_args[params[0]],templ_args[params[1]], wf_td.dt) 
    '''
    msk0 = np.logical_and((wl/f_mesh)<d_mesh, np.ones_like(f_mesh)*dt < 1/(f_mesh))
    #msk0 = wl*(1/f_mesh)<d_mesh
    msk1 = np.logical_not(msk0)
    return msk1




def white_noise_fd(F, magnitude=1.0e-22):
    '''
    Test function which creates constant noise curve power
    in the frequency domain.
    
    Input:
        F: 1D array, frequency region in which the noise should defined
        magnitude: float, default->1.0e-22
        
    Output:
        Power_spectrum object
    '''
    ASD_noise = np.sqrt(magnitude) * np.ones_like(F)
    return bs.Sensitivity_curve(F, ASD_noise, 'test_noise')

def find_postm(wf_obj, before=None, after=None, alpha=None):
    '''
    it crops the original waveform to just the postm part
    
    it computes the first local maximum of the
    instantaneous frequency as a function of time 
    as a way to define the beginning of the postmerger phase
    
    ''' 
    if before is None:
        before = 0
    
    if after is None:
        after = 0
     
        
    if alpha is None:
        alpha = 0.2
        
    md = wf_obj.mdta.copy()
    
    N_b = int(before/wf_obj.dt)
    N_a = int(after/wf_obj.dt)
        
    ind = np.argmax(wf_obj.GW_ampl)-N_b
    ind1 = np.argwhere(wf_obj.GW_ampl[ind:]>= 0.05*wf_obj.GW_ampl[ind:].max()).flatten()[-1] + N_a
    
    tukey  = signal.windows.tukey(wf_obj.td[ind:ind+ind1].size, alpha=alpha)
    wf = bs.waveform(wf_obj.td[ind:ind+ind1], wf_obj.hp[ind:ind+ind1]*tukey,wf_obj.hc[ind:ind+ind1]*tukey, md) 
    
    return wf



def find_h5_infd(path):
    directory = os.path.expanduser(path)
    file_ext = '**/*.h5'
    a = list(pathlib.Path(directory).glob(file_ext))
    for i in range(len(a)):
        a[i]=str(a[i])
    return a




def optimal_template(wf,sc,SNR, F, D, 
                     wl, tau, pfind,N,
                     th22max, cut_sc=None, 
                     plot=True, alpha=None,
                     phase_opt=True, zoom=None):
    
    if alpha is None:
        alpha=0
        
    if zoom is None:
        zoom = 0.4
    
    SNR1 = SNR.copy()
    mask = rjct_criteria_monochtempl(F, D, wf.dt,wl=wl)
    SNR1[mask] = np.nan
    SNR1[mask] = np.nan


    indT = np.unravel_index(np.nanargmax(SNR1), SNR1.shape)
    fmaxT, dmaxT, smaxT = F[indT], D[indT], SNR1[indT]
    
    
    u = monochr_templ(fmaxT, dmaxT, wf.dt, wf.t_0,
                      amp=1e-22, phase=0, alpha=alpha)
    S = snr.matched_filtering(wf.hp_get_tdobj(), u, sc)
    s = S.recovered_SNR3(N=N, cut_sc=cut_sc)
    sopt = np.abs(S.optimal_SNR(N=N, cut_sc=cut_sc).data).max()

    if pfind is not None:
        s = s.cut((pfind, s.td[-1]))
    ph = bs.time_domain(s.td, np.unwrap(np.angle(s.data)))
    val = tau
    if val is None:
        if phase_opt==True:
            ind = np.argmax(np.abs(s.data))
        elif phase_opt==False:
            ind = np.argmax(s.data.real)
    else:
        ind = np.where(s.td>=val)[0][0]

    phimax, taumax, smax = ph.data[ind], s.td[ind], s.data[ind]
    
    if phase_opt==True:
        assert np.isclose(smaxT,np.abs(smax))
    elif phase_opt==False:
        assert np.isclose(smaxT,smax.real)
    
    if plot==True:
        fig1 = plt.figure(dpi=900)
        im = plt.pcolormesh(F, D*1e3, SNR1/sopt,
                            shading='gouraud',cmap='cubehelix',
                           antialiased=True,rasterized=True)
        im.set_clim(0,1)
        fig1.colorbar(im, label=r'$\mathrm{\alpha=\frac{\rho_{rec}}{\rho_{opt}}}$', orientation='horizontal')
        #plt.title(f'SNR={smaxT:.3f} at f={fmaxT:.2f}[Hz], d={}')
        plt.scatter(fmaxT, dmaxT*1e3, marker='x', c='black', s=7)
        
        y = D.T[0][-1]*1e3*0.3
        x = F[0][-1]
        st = 'Best match found:\n'+f'f={fmaxT:.2f}[Hz]\n d={dmaxT*1e3:.2f}[ms]\n'+r'$\mathrm{{\tau}}$'+f'={taumax*1e3:.2f}[ms]\n'+r'$\phi$'+f'={phimax:.2f}[rad]\n'+r'$\alpha$'+f'={abs(smax)/sopt:.5f}'
        plt.text(1.05*x, y, st, fontsize=10)
        plt.xlabel('f [Hz]', fontsize=10)
        plt.ylabel('d [ms]',fontsize=10)
        
        plt.tight_layout()
        
        fig2 = plt.figure(dpi=900)
        plt.subplot(1,2,1)
        plt.title(f'SNRmax={np.abs(smax):.5f}')
        plt.plot(s.td*1e3, s.data.real)
        plt.plot(s.td*1e3, s.data.imag)
        plt.plot(s.td*1e3, np.abs(s.data))
        plt.ylabel(r'$\mathrm{\alpha=\frac{SNR_{rec}}{SNR_{opt}}}$')
        plt.xlabel(r'$\tau$ [ms]')
        plt.axhline(np.abs(smax), linestyle='dashed', color='r')
        plt.axvline(taumax*1e3, linestyle='dashed', color='r')


        plt.subplot(1,2,2)
        plt.title(f'taumax={taumax*1e3:.5f} [ms]')
        plt.plot(ph.td*1e3, ph.data)
        plt.axhline(phimax, linestyle='dashed', color='r')
        plt.axvline(taumax*1e3, linestyle='dashed', color='r')
        plt.ylabel(r'$\mathrm{\phi_0}$')
        plt.xlabel(r'$\tau$ ms')
       
        plt.tight_layout()
        
        u = monochr_templ(fmaxT, dmaxT, wf.dt, 
                     wf.t_0, amp=1.2*np.mean(wf.GW_ampl), phase=phimax, alpha=alpha)
        w = wf.hp_get_tdobj()
        Mf = snr.matched_filtering(w, u, sc)
        w1, u1 = Mf.length_matching2(N=N)

        m = wf.GW_ampl.max()
        fig3  = plt.figure(dpi=900)
        ax = fig3.add_subplot()
        md = wf.mdta.copy()
        name = md['name']
        #ax.set_title(name, fontsize=10)
        ax.text(0.7*wf.dur*1e3, -1.8*m,name, fontsize=10)
        ax.plot(w1.td*1e3, w1.data)
        ax.plot((u1.td+taumax)*1e3, u1.data)
        ax.yaxis.set_major_formatter(FormatStrFormatter('%1.0e'))

        

        m = wf.GW_ampl.max()
        ax.set_ylim(-2.*m, 2.*m)
        ax.set_xlim(-0.1*wf.dur*1e3, 1.1*wf.dur*1e3)
        ax.set_xlabel(r'$\mathrm{t-t_{merg}}$[ms]')
        ax.set_ylabel('strain @ 100.0 MPc')
        
        axins = ax.inset_axes([0.55, 0.7, 0.4, 0.2])
        axins.plot(w1.td*1e3, w1.data)
        axins.plot((u1.td+taumax)*1e3, u1.data)
        axins.set_xlim(-0.05*wf.dur*1e3, zoom*wf.dur*1e3)
        axins.set_ylim(-1.2*m, 1.2*m)

       
        
        axins.set_xticks([])
        axins.set_yticks([])
        
        ax.indicate_inset_zoom(axins, edgecolor="black")
        
        plt.tight_layout()
        
        return fig1, fig2, fig3, fmaxT, dmaxT, taumax, phimax, np.abs(smax), sopt
    else:
        return fmaxT, dmaxT, taumax, phimax, np.abs(smax), sopt



import numpy as np
import matplotlib.pyplot as plt
from matplotlib.patches import Rectangle
import matplotlib.image as mpimg
from mpl_toolkits.axes_grid1 import make_axes_locatable
import matplotlib.ticker

import base as bs
import utils as ut
import snr as snr

def wf_regions(td_obj, ax=None,**kwargs):
    '''
    this is function that draws square contours 
    to make easier, the regions of the timeseries
    on is taking into account
    
    input:
    
    td_obj  --> time_domain objects which packs time(x axis data)
                                         and strain(y axis data)
    ax      --> pyplot axes object as a container for the plot
    
    **kwargs--> section: part pf the plot one wants to crop
                         must be a tuple and it has two cases:
                         
                * to draw just one: 
                  section=('postm'/'prem'/'merg',
                            (t_init, t_fin))
                * to draw the three of them:
                  section=('all',((t_init, t_fin)  prem
                                  (t_init, t_fin)  merg 
                                  (t_init, t_fin)) postm
                                  
    example:
                         
        fig = plt.figure(figsize=(15,5))
        ax0 = fig.add_subplot(2,1,1)
        ax1 = fig.add_subplot(2,1,2)  
        wf_regions(wf0,ax=ax0, section=('merg', merg))
        wf_regions(wf0,ax=ax1, section=('all', (prem, merg, postm)))
        
    Output:
        pyplot axes object filled with a time series and the regions 
        marked with dashed squares
    
    
    '''
    height = np.max(td_obj.data)*(1+0.1)
    if ax is None:
        ax = plt.gca()
        
    ax.plot(td_obj.td*1e3, td_obj.data)

    if 'section' in kwargs:
        key = kwargs['section'][0]
        value = kwargs['section'][1]
        
        if key=='postm' or key=='prem' or key=='merg':
            ax.add_patch( Rectangle((value[0],-height),value[1]-value[0], 2*height,
                                    fc ='none', ec ='red', lw = 1, linestyle='--'))
            print(0)
        elif key=='all':
            color = ['red', 'green','orange']
            for i in range(len(value)):
                ax.add_patch( Rectangle((value[i][0],-height),value[i][1]-value[i][0], 2*height,
                        fc ='none', ec =color[i], lw = 1, linestyle='--'))
    return ax
    
    
def rjct_criteria_monochtempl(f_mesh,d_mesh,dt,wl=1):
    '''
    This funtion creates a mask wich rejects points in the 2D parameter space for 
    monochromatic wave templates. It does it based on:
        *Nyquist thorem, templates must have a sampling rate haigh enough to avoid 
                         aliasing
        *templates with less than one full cycle are rejected
    
    Input:
        f_mesh:2D array, frequency mesh
        d_mesh:2D array, duration mesh
        dt: float, sampling rate
    Output:
        msk1, mask(2D numpy array of bools) which is False at those rejected spots
              in the param space where one gets an unaccepted template
     
    Example:
        accepted_tmpl = rjct_criteria(templ_args[params[0]],templ_args[params[1]], wf_td.dt) 
    '''
    #msk0 = np.logical_and((1/f_mesh)<d_mesh, np.ones_like(f_mesh)*dt < 1/(*f_mesh))
    msk0 = wl*(1/f_mesh)<d_mesh
    msk1 = np.logical_not(msk0)
    return msk1
    
    
def best_SNR(B, wf, sc_ps,*templ_args,params=None,
             durfix=None,rjct_criteria=None, show=True):
    '''
    This is a plotter function that maps 2D surfaces embeded en a 3D space, where
    x,y & z coordinates are frequency,duration & SNR_rec
    
    Input:
    
        B: ND array, containing recovered SNR from a recursive matched filtering algorithm
           on a bank of templates which depend on *templ_args
           
        *templ_args: ND arrays, which are n-dimensional meshs, on which the template model
                     depends
        
        params: tuple, to indicate which of the N parameters are used in the 2D surface. By 
                default one should choose (0,1)
        
        rjct_criteria: ND mask, mask for rejecting points of parameter space which for some reason
                       are not wanted in the analysis
                
    Output:
        fig: contains all the plots 
        ind: tuple associated with the spot in which the maximum recovered SNR is 
        
    '''
    
    if params is None:
        params=(0,1)
    
    if rjct_criteria is not None:
        B[rjct_criteria]=np.nan
    

    if durfix is None:
        ind = np.unravel_index(np.nanargmax(B), B.shape)
    else:
        ind1 = np.argwhere(templ_args[params[1]]>=durfix).flatten()[0]
        ind2 = np.nanargmax(B[ind1,:])
        ind = (ind1, ind2)


    # this is for plotting pursposes, setting all the invalid values to zero SNR 
    if rjct_criteria is not None:
        B[rjct_criteria]= 0.0

    S_max = B[ind]
    par0_max = templ_args[params[0]][ind]
    par1_max = templ_args[params[1]][ind]

    temp = ut.monochr_templ(par0_max,par1_max,wf.dt, t_0=wf.t_0, amp=np.mean(wf.GW_ampl))

    uf = wf.hp_get_tdobj().to_fd()
    tf = temp.to_fd()   

    if show==True:
        fig = plt.figure(figsize=(7,4.5),constrained_layout=True)
        fig.subplots_adjust(wspace=5, hspace=0.8)
        #gs = fig.add_gridspec(8,28)    

        #ax0 = fig.add_subplot(gs[1:7,17:])
        #ax0.set_xlabel('f[Hz]', fontsize=15)
        #ax0.set_ylabel(r'$\mathrm{\sqrt{f}\cdot h(f)\;\;}$'+r'[$\mathrm{\frac{1}{\sqrt{Hz}}}$]', fontsize=15)
        ax1 = fig.add_subplot()#gs[:,:10])
        ax1.set_xlabel('$\mathrm{f[Hz]}$', fontsize=30)
        ax1.set_ylabel('$\mathrm{d[ms]}$', fontsize=30)

        im1 = ax1.pcolormesh(templ_args[params[0]],templ_args[params[1]]*1e3,B,cmap='cubehelix',shading='nearest',rasterized=True,antialiased=True, linewidth=0.0)
        cbar= fig.colorbar(im1, ax=ax1)
        cbar.set_label(r'$\mathrm{\rho}$', labelpad=10,rotation=0, fontsize=30)
        cbar.ax.tick_params(labelsize=20)
        ax1.scatter(par0_max,par1_max*1e3,c='red')
        ax1.axhline(par1_max*1e3, color='red', linestyle='dashed')
        ax1.tick_params(axis='both', which='major', color='black', labelsize=20)


        #ax0.plot(uf.fd, np.sqrt(np.abs(uf.fd))*np.abs(uf.data),label=r'$\mathrm{h_+ postm}$')
        #ax0.plot(tf.fd, np.sqrt(np.abs(tf.fd))*np.abs(tf.data),label='best template')
        #ax0.plot(sc_ps.fd, sc_ps.ASD, label='ASD ET')
        #ax0.set_xscale('log')
        #ax0.set_yscale('log')

        #ax0.set_xticks([20, 200, 2000, 5000, 10000])
        #ax0.get_xaxis().set_major_formatter(matplotlib.ticker.ScalarFormatter())
        #ax0.tick_params(axis='x', which='major', color='black', width=2, rotation=-45, length=10)

        #ax0.set_xlim(10, 12000)
        #ax0.set_ylim(0, 1e-20)
        #fig.legend(fontsize=12)
        plt.tight_layout()
    
    return par0_max, par1_max, S_max


def best_monochr(f, d, wf_td, sc_ps, max_templ_dur=None, taufix=None, phase_opt=True, show=True):
    '''
    This function takes the output of the best_SNR function, and shows several
    plots in the time domain and frequency domain for the case of monochromatic
    template models
    
    Input:
    
        mesh_ind: tuple, contains the position of the maximum
        f_mesh: 2D array, frequency mesh
        d_mesh: 2D array, duration mesh
        
        sc_ps: Sensitivity_curve object
               which hold the following information:
               self.fd-> non-equally spaced array
               self.ASD-> non-equally spaced array
               self.PSD-> non-equally spaced array
        
        wf_td: time_domain object
               which holds the following attributes
                    dt, sampling rate
                    dur, duration
                    t_0, initial phase
                    td, time array 
                    data, data array
         
        fig: pyplot figure object, default is None, if it is not the case, the user should pass
             a custom figure, e.g. fig=plt.figure(figsize=(15,5))
    
    Output:
        fig, contains all the plots
    '''
    phi = 0
    for i in range(2):
        tmpl_td = ut.monochr_templ(f, d, wf_td.dt, t_0=wf_td.t_0, phase=-phi, amp=np.sqrt(np.mean(wf_td.data**2)))
        if i==0:
            complex_SNR = (snr.matched_filtering(wf_td,tmpl_td, sc_ps)).recovered_SNR1(max_templ_dur=max_templ_dur)
            
            complex_SNR_s = complex_SNR[0]            
            complex_SNR_ph = complex_SNR[1]
            #complex_SNR_s = complex_SNR_s.cut((0, complex_SNR_s.td[-1]))
            #complex_SNR_ph = complex_SNR_ph.cut((0, complex_SNR_ph.td[-1]))            
            
            if taufix is None and phase_opt==True:
                ind = np.argmax(np.abs(complex_SNR_s.data))
                tau = complex_SNR_s.td[ind]
                s = np.abs(complex_SNR_s.data)[ind]
                phi = complex_SNR_ph.data[ind]

            elif taufix is None and phase_opt==False:
                ind = np.argmax(complex_SNR_s.data.real)
                tau = complex_SNR_s.td[ind]
                s = np.abs(complex_SNR_s.data)[ind]
                phi = complex_SNR_ph.data[ind]

            elif taufix is not None and phase_opt==True:
                ind = np.argwhere(complex_SNR_s.td>=taufix).flatten()[0]
                tau = complex_SNR_s.td[ind]
                s = np.abs(complex_SNR_s.data)[ind]
                phi = complex_SNR_ph.data[ind]

            elif taufix is not None and phase_opt==False:
                ind = np.argwhere(complex_SNR_s.td>=taufix).flatten()[0]
                tau = complex_SNR_s.td[ind]
                s = complex_SNR_s.data.real[ind]
                phi = complex_SNR_ph.data[ind]
    
    if show==True:
        fig = plt.figure(figsize=(7,4.5))

        gs = fig.add_gridspec(8,16)

        ax0 = fig.add_subplot(gs[:,:])

        ax0.set_ylabel('Strain at 100 Mpc', fontsize=30)
        ax0.set_xlabel('$\mathrm{t-t_{merger}[ms]}$', fontsize=30)
        ax0.plot((wf_td.td-wf_td.t_0)*1e3,wf_td.data,label=r'$\mathrm{h_{+}}$ postmerger')
        ax0.plot((tmpl_td.td+tau-wf_td.t_0)*1e3,tmpl_td.data, linestyle='dashed', label='best template')
        ax0.legend(loc='upper right',fontsize=20)
        plt.tight_layout()
    return  f, d, s, tau, phi  
 
    
    
    



import numpy as np
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
from matplotlib.gridspec import GridSpec
from matplotlib.ticker import FormatStrFormatter
import ray

def catalog_analysis(waveform_bank, func, **func_kwargs):
    res = [func(wf_id, **func_kwargs) for wf_id in waveform_bank]
    res = {list(x.keys())[0]:x[list(x.keys())[0]] for x in res}
    return res

def par_catalog_analysis(waveform_bank, func, **func_kwargs):
    res = ray.get([func.remote(wf_id, **func_kwargs) for wf_id in waveform_bank])
    res = {list(x.keys())[0]:x[list(x.keys())[0]] for x in res}
    return res


def relabler_cat(A, input_string, maxdigs=None):
    '''
    relabeling function, 
    For those cases where the names of simulations
    on catalogs are ridiculously long
    
    Input: 
        A, dict, original dictionary
        input_tring: first characters of the key, e.g
                     lal ---> lal001
                         ---> lal002
                         ---> ...
        maxdigs:int, number of digits after the
                input string. e.g.
                3--> lal001
                6--> lal000001
    Output:
    
        C, dict. New relabled dictionary
        D, dict, small dictionary mapping old and new keys
                
    '''
    
    
    B = A.copy()
    C = {}
    D = {}
    if maxdigs is None:
        maxdigs = 4
    
    for i, key in enumerate(B.keys()):
        i+=1
        st = input_string + str(i).rjust(maxdigs, '0')
        D[st] = key
        C[st] = B[key]
    return C, D

def scatter_cat(A, func1, func2, var1, var2,ax=None,
                bounds=None, boundsv=None, mdta=None):
    '''
    scatter plot with highlighted regions 
    from results dictionary.
    
    Input:
        A: dict. results from the catalog analysis
        func1&2: ufunc, defines the values of each axis 
        var1&2: str, keys to values in the dictionary
        
        bounds: list of lists, defines the boundaries 
                each region will be highlighted with
                different colors.e.g.
                
                bounds=[[x1,x2, y1,y2],   -> bound region 1
                        [x1,x2, y1,y2],   -> bound region 2
                        ...,
                        ]
                        
        boundsv: list, highligh according to metadata.e.g.
                boundsv= ['mass-ratio', 1.1,  2.1] 
        
    Output:
        ax: pyplot axes
        R: list of waveforms lying on the highlighted regions.
    
    '''
    
    if mdta is None:
        mdta = False
    

    paramsx0 = lambda key: (A[key]['mdta'][param][0] for param in var1)
    paramsx1 = lambda key: (A[key][param][0] for param in var1)

        

    paramsy0 = lambda key: (A[key]['mdta'][param][0] for param in var2)
    paramsy1 = lambda key: (A[key][param][0] for param in var2)    
     
    
    if ax is None:
        fig = plt.figure(dpi=1200)
        ax = fig.add_subplot()
    
    for key in A.keys():
        
        try:
            x = func1(*paramsx0(key))
        except:
            x = func1(*paramsx1(key))
            
        try:
            y = func2(*paramsy0(key))
        except:
            y = func2(*paramsy1(key))
            
        ax.scatter(x, y, color='blue', s=1)
        
        
    if bounds is not None:
        c = ['r', 'darkgreen', 'black', 'orange']
        R = {'r':[], 'darkgreen':[], 'black':[], 'orange':[]}
        for key in A.keys():
            try:
                x = func1(*paramsx0(key))
            except:
                x = func1(*paramsx1(key))

            try:
                y = func2(*paramsy0(key))
            except:
                y = func2(*paramsy1(key))
                
            for i in range(len(bounds)):
                if x>bounds[i][0] and x<bounds[i][1] and y>bounds[i][2] and y<bounds[i][3]:
                    ax.scatter(x, y, marker='x', color=c[i], s=15)
                    R[c[i]].append(key)
        return ax, R           
    
    if boundsv is not None:
        R = {'r':[]}
        for key in A.keys():
            var = boundsv[0]
            
            try:
                x = func1(*paramsx0(key))
            except:
                x = func1(*paramsx1(key))

            try:
                y = func2(*paramsy0(key))
            except:
                y = func2(*paramsy1(key))
            
            
            z = A[key]['mdta'][var][0]
            
            if z>boundsv[1] and z<=boundsv[2] :
                ax.scatter(x, y, marker='x', color='r', s=15)
                R['r'].append(key)
        return ax, R
    
    return ax, {'blue':list(A.keys())}
      

def hist_cat(A, L, g, ax=None,var=None, varmdta=None, bins=None):
    
    if ax is None:
        fig = plt.figure(dpi=1200)
        ax = fig.add_subplot()
    
    
    
    def extract(A, key, var=None, varmdta=None):
        if varmdta is None:
            gen = (A[key][param][0] for param in var)
            res = g(*gen)
        else:
            gen = (A[key]['mdta'][param][0] for param in varmdta)
            res = g(*gen)
        return res
    
 
    f = lambda key: extract(A, key, var=var, varmdta=varmdta)
    
    
    res = [[*map(f, row)] for row in list(L.values())]
    
    #print(res)

    
    ax.hist(res, color=list(L.keys()) , stacked=True, rwidth=0.9, bins=bins)
    
    return ax
    



def layer_cat(layers, A, mdta = False, 
              show_eos=True ,eos_palette=None, step=4):
    

    
    eos = set( [A[name]['mdta']['eos'][0]  for name in A] )
    eos = {name:eos_palette[i] for i, name in enumerate(eos)}
    
    fig = plt.figure(figsize=(30,20))#, dpi=1200)
    n = len(layers)
    gs = GridSpec(n,9, figure=fig)
    if show_eos==True:
        axes = [ fig.add_subplot(gs[i-1, :8]) if i<n+1 else fig.add_subplot(gs[:,8: ]) for i in range(1,n+2)]


    else:
        axes = [ fig.add_subplot(gs[i-1, :]) for i in range(1,n+1)]
        [ax.tick_params(axis='x', which='both', bottom=False, top=False, labelbottom=False) for i,ax in enumerate(axes) if i<n-1]

    
    for j, layer in enumerate(layers):
        L, V, C = ['name' for x in A.keys()], [0 for x in A.keys()], ['color' for x in A.keys()]
        for i, name in enumerate(A.keys()):
            eos_name = A[name]['mdta']['eos'][0]
            #order = int(name.split(':')[-1])
            
            if mdta==False:
                value = A[name][layer][0]
            else:
                value = A[name]['mdta'][layer][0]
                
            c = eos[eos_name]
            #print(order-1)
            
            L[i] = name
            V[i] = value
            C[i] = c
            
        bot, top = np.nanmin(V), np.nanmax(V)
        epsilon =0.2*(0.5 * (bot+top))
        V = [top*10 if np.isnan(x) else x for x in V]
        #print(len(C), len(V))
        axes[j].scatter(L, V, c=C, marker='x', s=40)
        axes[j].set_xticks(np.arange(len(L)), L, rotation=90)
        axes[j].set_yticks(np.linspace(bot-epsilon, top+epsilon, 5))
        
        axes[j].tick_params(axis='x',labelsize=40)
        axes[j].tick_params(axis='y',labelsize=30)

        

        axes[j].xaxis.set_major_locator(ticker.MultipleLocator(step))
        axes[j].yaxis.set_major_formatter(FormatStrFormatter('%1.1e'))
        axes[j].grid()
        
        axes[j].set_ylim(bot-epsilon, top+epsilon)

        
    axes[-1].set_xticks([])
    axes[-1].set_yticks([])

    [ax.set_xticklabels([]) for i,ax in enumerate(axes) if i<n-1]

    [axes[-1].scatter(0,0, color=eos[name], marker='x', label=name, s=100) for name in eos.keys()]
    axes[-1].set_xlim(12, 100)
    axes[-1].set_ylim(12, 100)


    axes[-1].legend(loc='center', fontsize=5, prop={'size': 35})
    axes[-1].axis('off')
    
    plt.tight_layout()
        
            

    return axes